﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.GroupingVector {
  [Item("SingleGroupingMoveSoftTabuCriterion", "<Missing>")]
  [StorableClass]
  public class SingleGroupingMoveSoftTabuCriterion : SingleSuccessorOperator, IGroupingVectorMoveOperator, ITabuChecker {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<SingleGroupingMove> SingleGroupingMoveParameter {
      get { return (ILookupParameter<SingleGroupingMove>)Parameters["SingleGroupingMove"]; }
    }
    public ILookupParameter<GroupingVectorEncoding> GroupingVectorParameter {
      get { return (ILookupParameter<GroupingVectorEncoding>)Parameters["GroupingVector"]; }
    }
    public ILookupParameter<ItemList<IItem>> TabuListParameter {
      get { return (ILookupParameter<ItemList<IItem>>)Parameters["TabuList"]; }
    }
    public ILookupParameter<BoolValue> MoveTabuParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["MoveTabu"]; }
    }
    public IValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    public ValueParameter<BoolValue> UseAspirationCriterionParameter {
      get { return (ValueParameter<BoolValue>)Parameters["UseAspirationCriterion"]; }
    }

    [StorableConstructor]
    protected SingleGroupingMoveSoftTabuCriterion(bool deserializing) : base(deserializing) { }
    protected SingleGroupingMoveSoftTabuCriterion(SingleGroupingMoveSoftTabuCriterion original, Cloner cloner) : base(original, cloner) { }
    public SingleGroupingMoveSoftTabuCriterion()
      : base() {
      Parameters.Add(new LookupParameter<SingleGroupingMove>("SingleGroupingMove", "The move to evaluate."));
      Parameters.Add(new LookupParameter<GroupingVectorEncoding>("GroupingVector", "The solution to evaluate."));
      Parameters.Add(new LookupParameter<BoolValue>("MoveTabu", "The variable to store if a move was tabu."));
      Parameters.Add(new LookupParameter<ItemList<IItem>>("TabuList", "The tabu list."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, else if it is a minimization problem."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The quality of the current move."));
      Parameters.Add(new ValueParameter<BoolValue>("UseAspirationCriterion", "Whether to use the aspiration criterion or not.", new BoolValue(true)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SingleGroupingMoveSoftTabuCriterion(this, cloner);
    }

    public override IOperation Apply() {
      ItemList<IItem> tabuList = TabuListParameter.ActualValue;
      SingleGroupingMove move = SingleGroupingMoveParameter.ActualValue;
      GroupingVectorEncoding groupingVector = GroupingVectorParameter.ActualValue;
      int length = groupingVector.GroupingVector.Length;
      double moveQuality = MoveQualityParameter.ActualValue.Value;
      bool maximization = MaximizationParameter.ActualValue.Value;   
      bool useAspiration = UseAspirationCriterionParameter.Value.Value;
      bool isTabu = IsMoveTabu(tabuList, move, groupingVector, moveQuality, maximization, useAspiration);

      MoveTabuParameter.ActualValue = new BoolValue(isTabu);
      return base.Apply();
    }

    private bool IsMoveTabu(ItemList<IItem> tabuList, SingleGroupingMove move, GroupingVectorEncoding groupingVector, double moveQuality, bool maximization, bool useAspiration) {
      var movedGroupingVector = new IntegerVector(groupingVector.GroupingVector);
      movedGroupingVector[move.Index] = move.NewGroup;
      foreach (IItem tabuMove in tabuList) {
        SingleGroupingMoveAttribute sgmAttr = (tabuMove as SingleGroupingMoveAttribute);
        if (sgmAttr != null
          && (!useAspiration
              || maximization && moveQuality <= sgmAttr.MoveQuality
              || !maximization && moveQuality >= sgmAttr.MoveQuality)) {
                if (move.Index == sgmAttr.ItemIndex && groupingVector.GroupingVector[move.Index] == sgmAttr.AssignedBin)
                  return true;
        }
      }
      return false;
    }
  }
}
