﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.MultiComponentVector {
  [Item("MultiComponentVectorMoveHardTabuCriterion", "<Missing>")]
  [StorableClass]
  public class MultiComponentVectorMoveHardTabuCriterion : SingleSuccessorOperator, IMultiComponentVectorMoveOperator, ITabuChecker {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<MultiComponentVectorMove> MultiComponentVectorMoveParameter {
      get { return (ILookupParameter<MultiComponentVectorMove>)Parameters["MultiComponentVectorMove"]; }
    }
    public ILookupParameter<MultiComponentVectorEncoding> MultiComponentVectorParameter {
      get { return (ILookupParameter<MultiComponentVectorEncoding>)Parameters["MultiComponentVector"]; }
    }
    public ILookupParameter<ItemList<IItem>> TabuListParameter {
      get { return (ILookupParameter<ItemList<IItem>>)Parameters["TabuList"]; }
    }
    public ILookupParameter<BoolValue> MoveTabuParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["MoveTabu"]; }
    }
    public IValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    public ValueParameter<BoolValue> UseAspirationCriterionParameter {
      get { return (ValueParameter<BoolValue>)Parameters["UseAspirationCriterion"]; }
    }

    [StorableConstructor]
    protected MultiComponentVectorMoveHardTabuCriterion(bool deserializing) : base(deserializing) { }
    protected MultiComponentVectorMoveHardTabuCriterion(MultiComponentVectorMoveHardTabuCriterion original, Cloner cloner) : base(original, cloner) { }
    public MultiComponentVectorMoveHardTabuCriterion()
      : base() {
        Parameters.Add(new LookupParameter<MultiComponentVectorMove>("MultiComponentVectorMove", "The move to evaluate."));
      Parameters.Add(new LookupParameter<MultiComponentVectorEncoding>("MultiComponentVector", "The solution to evaluate."));
      Parameters.Add(new LookupParameter<BoolValue>("MoveTabu", "The variable to store if a move was tabu."));
      Parameters.Add(new LookupParameter<ItemList<IItem>>("TabuList", "The tabu list."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, else if it is a minimization problem."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The quality of the current move."));
      Parameters.Add(new ValueParameter<BoolValue>("UseAspirationCriterion", "Whether to use the aspiration criterion or not.", new BoolValue(true)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MultiComponentVectorMoveHardTabuCriterion(this, cloner);
    }

    public override IOperation Apply() {
      ItemList<IItem> tabuList = TabuListParameter.ActualValue;
      MultiComponentVectorMove move = MultiComponentVectorMoveParameter.ActualValue;
      MultiComponentVectorEncoding multiComponentVector = MultiComponentVectorParameter.ActualValue;
      int length = multiComponentVector.PackingInformations.Count;
      double moveQuality = MoveQualityParameter.ActualValue.Value;
      bool maximization = MaximizationParameter.ActualValue.Value;
      bool useAspiration = UseAspirationCriterionParameter.Value.Value;
      bool isTabu = IsMoveTabu(tabuList, move, multiComponentVector, moveQuality, maximization, useAspiration);

      MoveTabuParameter.ActualValue = new BoolValue(isTabu);
      return base.Apply();
    }

    private bool IsMoveTabu(ItemList<IItem> tabuList, MultiComponentVectorMove move, MultiComponentVectorEncoding multiComponentVector, double moveQuality, bool maximization, bool useAspiration) {
      bool isTabu = false;
      SingleItemRotationMove sirm = move as SingleItemRotationMove;
      SingleGroupingMove sgm = move as SingleGroupingMove;
      SwapPositionMove spm = move as SwapPositionMove;
      ChangePositionMove cpm = move as ChangePositionMove;
      MCVTripleMove tm = move as MCVTripleMove;
      var attributes = tabuList.FindAll(ta => ta.GetType().Equals(move.GetMoveAttributeType()));

      if (sirm != null) {
        foreach (IItem tabuMove in attributes) {
          SingleItemRotationMoveAttribute sirmAttr = (tabuMove as SingleItemRotationMoveAttribute);
          if (sirmAttr.MoveQuality.Equals(moveQuality))
            return true;
          if (sirmAttr != null
            && (!useAspiration
                || maximization && moveQuality <= sirmAttr.MoveQuality
                || !maximization && moveQuality >= sirmAttr.MoveQuality)) {
                  if (sirm.ItemIndex == sirmAttr.ItemIndex || multiComponentVector.PackingInformations[sirm.ItemIndex].Rotated == sirmAttr.ItemRotation)
                    isTabu = true;
          }
          if (isTabu) break;
        }
      } else if (sgm != null) {
        foreach (IItem tabuMove in attributes) {
          SingleGroupingMoveAttribute sgmAttr = (tabuMove as SingleGroupingMoveAttribute);
          if (sgmAttr.MoveQuality.Equals(moveQuality))
            return true;
          if (sgmAttr != null
            && (!useAspiration
                || maximization && moveQuality <= sgmAttr.MoveQuality
                || !maximization && moveQuality >= sgmAttr.MoveQuality)) {
                  if ( multiComponentVector.PackingInformations[sgm.Index].ItemIndex ==  multiComponentVector.PackingInformations[sgmAttr.Index].ItemIndex
                    || multiComponentVector.PackingInformations[sgm.Index].AssignedBin == sgmAttr.AssignedBin)
                    isTabu = true;
          }
          if (isTabu) break;
        }
      } else if (spm != null) {
        foreach (IItem tabuMove in attributes) {
          SwapPositionMoveAttribute spmAttr = (tabuMove as SwapPositionMoveAttribute);
          if (spmAttr.MoveQuality.Equals(moveQuality))
            return true;
          if (spmAttr != null
            && (!useAspiration
                || maximization && moveQuality <= spmAttr.MoveQuality
                || !maximization && moveQuality >= spmAttr.MoveQuality)) {
            if (spm.Index1 == spmAttr.Index1
              || spm.Index2 == spmAttr.Index2
              || multiComponentVector.PackingInformations[spm.Index1].ItemIndex == spmAttr.ItemIndex1
              || multiComponentVector.PackingInformations[spm.Index2].ItemIndex == spmAttr.ItemIndex2)
              isTabu = true;
          }
          if (isTabu) break;
        }
      } else if (cpm != null) {
        foreach (IItem tabuMove in attributes) {
          ChangePositionMoveAttribute cpmAttr = (tabuMove as ChangePositionMoveAttribute);
          if (cpmAttr.MoveQuality.Equals(moveQuality))
            return true;
          if (cpmAttr != null
            && (!useAspiration
                || maximization && moveQuality <= cpmAttr.MoveQuality
                || !maximization && moveQuality >= cpmAttr.MoveQuality)) {
            if (cpm.Index == cpmAttr.Index
        || cpm.TargetIndex == cpmAttr.TargetIndex
        || multiComponentVector.PackingInformations[cpm.Index].ItemIndex == cpmAttr.ItemIndex)
              isTabu = true;
          }
          if (isTabu) break;
        }
      } else if (tm != null) {
        foreach (IItem tabuMove in attributes) {
          MCVTripleMoveAttribute tmAttr = (tabuMove as MCVTripleMoveAttribute);
          if (tmAttr.MoveQuality.Equals(moveQuality))
            return true;
          if (tmAttr != null
            && (!useAspiration
                || maximization && moveQuality <= tmAttr.MoveQuality
                || !maximization && moveQuality >= tmAttr.MoveQuality)) {
                  if ((tm.Index == tmAttr.Index && tm.TargetIndex == tmAttr.TargetIndex)
                      || tm.Index == tmAttr.Index
                      || multiComponentVector.PackingInformations[tm.Index].ItemIndex == tmAttr.ItemIndex
                    )
              isTabu = true;
          }
          if (isTabu) break;
        }
      } 

      

      return isTabu;
    }
  }
}
