﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.MultiComponentVector {
  [Item("Partially Matched Multi Component Vector Crossover", "An operator which crosses two MultiComponentVector representations.")]
  [StorableClass]
  public class PartiallyMatchedMultiComponentVectorCrossover :  MultiComponentVectorCrossover {
    [StorableConstructor]
    protected PartiallyMatchedMultiComponentVectorCrossover(bool deserializing) : base(deserializing) { }
    protected PartiallyMatchedMultiComponentVectorCrossover(PartiallyMatchedMultiComponentVectorCrossover original, Cloner cloner) : base(original, cloner) { }
    public PartiallyMatchedMultiComponentVectorCrossover()
      : base() {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PartiallyMatchedMultiComponentVectorCrossover(this, cloner);
    }

    public override MultiComponentVectorEncoding Cross(IRandom random, MultiComponentVectorEncoding parent1, MultiComponentVectorEncoding parent2) {
      if (parent1.PackingInformations.Count != parent2.PackingInformations.Count) throw new ArgumentException("PartiallyMatchedCrossover: The parent permutations are of unequal length.");
      if (parent1.PackingInformations.Count < 4) throw new ArgumentException("PartiallyMatchedCrossover: The parent permutation must be at least of size 4.");
      int length = parent1.PackingInformations.Count;
      PackingInformation[] result = new PackingInformation[length];
      int[] invResult = new int[length];

      int breakPoint1, breakPoint2;
      do {
        breakPoint1 = random.Next(length - 1);
        breakPoint2 = random.Next(breakPoint1 + 1, length);
      } while (breakPoint2 - breakPoint1 >= length - 2); // prevent the case [0,length-1) -> clone of parent1

      // clone parent2 and calculate inverse permutation (number -> index)
      try {
        for (int j = 0; j < length; j++) {
          result[j] = parent2.PackingInformations[j];
          invResult[result[j].ItemIndex] = j;
        }
      }
      catch (IndexOutOfRangeException) {
        throw new InvalidOperationException("PartiallyMatchedCrossover: The permutation must consist of consecutive numbers from 0 to N-1 with N = length of the permutation.");
      }

      for (int j = breakPoint1; j <= breakPoint2; j++) {
        PackingInformation orig = new PackingInformation (result[j]); // save the former value
        result[j] = parent1.PackingInformations[j]; // overwrite the former value with the new value
        int index = invResult[result[j].ItemIndex]; // look where the new value is in the offspring
        result[index] = orig; // write the former value to this position
        invResult[orig.ItemIndex] = index; // update the inverse mapping
        // it's not necessary to do 'invResult[result[j]] = j' as this will not be needed again
      }

      var childSolution = new MultiComponentVectorEncoding();
      childSolution.PackingInformations = new ItemList<PackingInformation> (result);
      return childSolution; //new Permutation(parent1.PermutationType, result);
    }

  }
}
