﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.Potvin {
  [Item("Potvin Encoding", "Represents an encoding for a bin packing problem using a delimited permutation of item-indexes.")]
  [StorableClass]
  public class PotvinEncoding : Item, IPackingSolutionEncoding{

    [Storable]
    public IntegerVector IntVector { get; set; }
    [Storable]
    public bool DelimiterInitialized { get; set; }

    private int nrOfBins = -1;
    public int NrOfBins { 
      get { 
        if (nrOfBins < 0) { 
          nrOfBins = 1;
          foreach (int i in IntVector) {
            if (i == -1)
              nrOfBins++;
          }
        }
        return nrOfBins;
      } 
    }

    public List<int> GetItemsInBin(int binNr) {
      int binCount = 0;
      List<int> result = new List<int> ();
      foreach (int i in IntVector) {
        if (i == -1) {
          binCount++;
          if (binCount > binNr)
            break;
        }
        if (binCount == binNr)
          result.Add(i);
      }
      return result;
    }

    [StorableConstructor]
    protected PotvinEncoding(bool deserializing) : base(deserializing) { }
    protected PotvinEncoding(PotvinEncoding original, Cloner cloner)
      : base(original, cloner) {
        this.IntVector = cloner.Clone(original.IntVector);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PotvinEncoding(this, cloner);
    }

    public PotvinEncoding()
      : base() {
        IntVector = new IntegerVector();
        DelimiterInitialized = false;
    }

    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append("[ ");
      sb.Append(IntVector.ToString());         
      sb.Append("]");
      return sb.ToString();
    }


    public override bool Equals(object obj) {
      if (obj.GetType() == typeof(PotvinEncoding))
        return AreEqual(this, obj as PotvinEncoding);

      return false;
    }
    public override int GetHashCode() {
      return IntVector.GetHashCode();
    }
    private static bool AreEqual(PotvinEncoding PotvinEncoding1, PotvinEncoding PotvinEncoding2) {
      return AreEqual(PotvinEncoding1.IntVector, PotvinEncoding2.IntVector);
    }

    private static bool AreEqual(IntArray v1, IntArray v2) {
      if (v1.Length != v2.Length)
        return false;
      for (int i = 0; i < v1.Length; i++) {
        if (v1[i] != v2[i])
          return false;
      }
      return true;
    }
  }
}
