﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.Potvin {
  [Item("Potvin Encoding Manipulator", "An operator which manipulates a Potvin representation.")]
  [StorableClass]
  public class PotvinManipulator : PackingSolutionManipulator, IPotvinOperator {
    [StorableConstructor]
    protected PotvinManipulator(bool deserializing) : base(deserializing) { }
    protected PotvinManipulator(PotvinManipulator original, Cloner cloner) : base(original, cloner) { }
    public PotvinManipulator()
      : base() {
        EncodedSolutionParameter.ActualName = "Potvin";
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PotvinManipulator(this, cloner);
    }

    protected int SelectRandomTourBiasedByLength(IRandom random, PotvinEncoding individual) {
      int tourIndex = -1;

      double sum = 0.0;
      double[] probabilities = new double[individual.Tours.Count];
      for (int i = 0; i < individual.Tours.Count; i++) {
        probabilities[i] = 1.0 / ((double)individual.Tours[i].Stops.Count / (double)ProblemInstance.Cities.Value);
        sum += probabilities[i];
      }

      for (int i = 0; i < probabilities.Length; i++)
        probabilities[i] = probabilities[i] / sum;

      double rand = random.NextDouble();
      double cumulatedProbabilities = 0.0;
      int index = 0;
      while (tourIndex == -1 && index < probabilities.Length) {
        if (cumulatedProbabilities <= rand && rand <= cumulatedProbabilities + probabilities[index])
          tourIndex = index;

        cumulatedProbabilities += probabilities[index];
        index++;
      }

      return tourIndex;
    }

    protected bool FindInsertionPlace(PotvinEncoding individual, int city, int routeToAvoid, bool allowInfeasible, out int route, out int place) {
      return individual.FindInsertionPlace(
        city, routeToAvoid, allowInfeasible, out route, out place);
    }

    protected void Manipulate(IRandom random, PotvinEncoding individual) {

      int selectedIndex = SelectRandomTourBiasedByLength(random, individual);
      if (selectedIndex >= 0) {
        Tour route1 =
          individual.Tours[selectedIndex];

        int count = route1.Stops.Count;
        int i = 0;
        while (i < count) {
          int insertedRoute, insertedPlace;

          int city = route1.Stops[i];
          route1.Stops.Remove(city);

          if (FindInsertionPlace(individual, city, selectedIndex, allowInfeasible, out insertedRoute, out insertedPlace)) {
            individual.Tours[insertedRoute].Stops.Insert(insertedPlace, city);
          } else {
            route1.Stops.Insert(i, city);
            i++;
          }

          count = route1.Stops.Count;
        }
      }
    }

    public override IOperation Apply() {
      var solution = EncodedSolutionParameter.ActualValue as PotvinEncoding;
      if (solution == null) throw new InvalidOperationException("PackingSolutionEncoding was not found or is not of type PotvinEncoding.");
      Manipulate(RandomParameter.ActualValue, solution);
      return base.Apply();
    }

  }
}
